'' This is a minimal version of an I2C driver in SPIN.  It assumes
'' that the SDA pin is one higher than the SCL pin.  It assumes that
'' neither the SDA nor the SCL pins have pullups, so drives both.
CON
   i2cACK    = 0                       ' I2C Acknowledge
   i2cNAK    = 1                       ' I2C No Acknowledge
   i2cXmit   = 0                       ' I2C Direction Transmit
   i2cRecv   = 1                       ' I2C Direction Receive
   i2cBoot   = 28                      ' I2C Boot EEPROM SCL Pin
   i2cEEPROM = $A0                ' I2C EEPROM Device Address

PUB i2cStart(i2cSCL) | i2cSDA   ' SDA goes HIGH to LOW with SCL HIGH
   i2cSDA := i2cSCL + 1
   outa[i2cSCL]~~               ' Initially drive SCL HIGH
   dira[i2cSCL]~~
   outa[i2cSDA]~~               ' Initially drive SDA HIGH
   dira[i2cSDA]~~
   outa[i2cSDA]~                ' Now drive SDA LOW
   outa[i2cSCL] ~               ' Leave SCL LOW
  
PUB i2cStop(i2cSCL) | i2cSDA    ' SDA goes LOW to HIGH with SCL High
   i2cSDA := i2cSCL + 1
   outa[i2cSCL] ~~              ' Drive SCL HIGH
   outa[i2cSDA] ~~              '  then SDA HIGH
   dira[i2cSCL] ~               ' Now let them float
   dira[i2cSDA] ~               ' If pullups present, they'll stay HIGH

PUB i2cWrite(i2cSCL, i2cData) : ackbit | i2cSDA
' Write i2c data.  Data byte is output MSB first, SDA data line is valid
' only while the SCL line is HIGH.  Data is always 8 bits (+ ACK/NAK).
' SDA is assumed LOW and SCL and SDA are both left in the LOW state.
   i2cSDA := i2cSCL + 1
   ackbit := 0 
   i2cData <<= 24
   repeat 8                      ' Output data to SDA
      outa[i2cSDA] := (i2cData <-= 1) & 1
      outa[i2cSCL] ~~            ' Toggle SCL from LOW to HIGH to LOW
      outa[i2cSCL] ~
   dira[i2cSDA] ~                ' Set SDA to input for ACK/NAK
   outa[i2cSCL] ~~
   ackbit := ina[i2cSDA]         ' Sample SDA when SCL is HIGH
   outa[i2cSCL] ~
   outa[i2cSDA] ~                ' Leave SDA driven LOW
   dira[i2cSDA] ~~

PUB i2cRead(i2cSCL, ackbit): i2cData | i2cSDA
' Read in i2c data, Data byte is output MSB first, SDA data line is
' valid only while the SCL line is HIGH.  SCL and SDA left in LOW state.
   i2cSDA := i2cSCL + 1
   i2cData := 0
   dira[i2cSDA]~                 ' Make SDA an input
   repeat 8                      ' Receive data from SDA
      outa[i2cSCL] ~~            ' Sample SDA when SCL is HIGH
      i2cData := (i2cData << 1) | ina[i2cSDA]
      outa[i2cSCL] ~
   outa[i2cSDA] := ackbit        ' Output ACK/NAK to SDA
   dira[i2cSDA] ~~
   outa[i2cSCL] ~~               ' Toggle SCL from LOW to HIGH to LOW
   outa[i2cSCL] ~
   outa[i2cSDA] ~                ' Leave SDA driven LOW
  

PUB i2cReadPage(i2cSCL, i2cAddr, addrReg, dataPtr, count) : ackbit
' Read in a block of i2c data.  Device address is i2cAddr.  Device starting
' address is addrReg.  Data address is at dataPtr.  Number of bytes is count.
' The device address is modified using the upper 3 bits of the 19 bit addrReg.
   i2cAddr |= addrReg >> 15 & %1110
   i2cStart(i2cSCL)
   ackbit := i2cWrite(i2cSCL, i2cAddr | i2cXmit)
   ackbit := (ackbit << 1) | i2cWrite(i2cSCL, addrReg >> 8 & $FF)
   ackbit := (ackbit << 1) | i2cWrite(i2cSCL, addrReg & $FF)          
   i2cStart(i2cSCL)
   ackbit := (ackbit << 1) | i2cWrite(i2cSCL, i2cAddr | i2cRecv)
   repeat count - 1
      byte[dataPtr++] := i2cRead(i2cSCL, i2cACK)
   byte[dataPtr++] := i2cRead(i2cSCL, i2cNAK)
   i2cStop(i2cSCL)
   return ackbit

PUB i2cWritePage(i2cSCL, i2cAddr, addrReg, dataPtr, count) : ackbit
' Write out a block of i2c data.  Device address is i2cAddr.  Device starting
' address is addrReg.  Data address is at dataPtr.  Number of bytes is count.
' The device address is modified using the upper 3 bits of the 19 bit addrReg.
' Most devices have a page size of at least 32 bytes, some as large as 128.
   i2cAddr |= addrReg >> 15 & %1110
   i2cStart(i2cSCL)
   ackbit := i2cWrite(i2cSCL, i2cAddr | i2cXmit)
   ackbit := (ackbit << 1) | i2cWrite(i2cSCL, addrReg >> 8 & $FF)
   ackbit := (ackbit << 1) | i2cWrite(i2cSCL, addrReg & $FF)          
   repeat count
      ackbit := ackbit << 1 | ackbit & $80000000 ' "Sticky" sign bit         
      ackbit |= i2cWrite(i2cSCL, byte[dataPtr++])
   i2cStop(i2cSCL)
   return ackbit

PUB i2cWriteWait(i2cSCL, i2cAddr, addrReg) : ackbit
' Wait for a previous write to complete.  Device address is i2cAddr.  Device
' starting address is addrReg.  The device will not respond if it is busy.
' The device address is modified using the upper 3 bits of the 18 bit addrReg.
   i2cAddr |= addrReg >> 15 & %1110
   i2cStart(i2cSCL)
   ackbit := i2cWrite(i2cSCL, i2cAddr | i2cXmit)
   i2cStop(i2cSCL)
   return ackbit